'use strict';

var multiformats = require('multiformats');
var writer = require('../lib/writer.js');
var common = require('./common.js');

const {toHex} = multiformats.bytes;
function collector(iterable) {
  return (async () => {
    const chunks = [];
    let length = 0;
    for await (const chunk of iterable) {
      chunks.push(chunk);
      length += chunk.length;
    }
    const bytes = new Uint8Array(length);
    length = 0;
    for (const chunk of chunks) {
      bytes.set(chunk, length);
      length += chunk.length;
    }
    return bytes;
  })();
}
describe('CarWriter', () => {
  let cborBlocks;
  let allBlocksFlattened;
  let roots;
  function assertCarData(actual) {
    common.assert.strictEqual(toHex(actual), toHex(common.carBytes), 'got expected bytes');
  }
  before(async () => {
    const data = await common.makeData();
    cborBlocks = data.cborBlocks;
    allBlocksFlattened = data.allBlocksFlattened;
    roots = [
      cborBlocks[0].cid,
      cborBlocks[1].cid
    ];
  });
  it('complete', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    common.assert.strictEqual(typeof out[Symbol.asyncIterator], 'function');
    const collection = collector(out);
    const writeQueue = [];
    for (const block of allBlocksFlattened) {
      writeQueue.push(writer$1.put(block));
    }
    writeQueue.push(writer$1.close());
    let collected = false;
    collection.then(bytes => {
      collected = true;
      assertCarData(bytes);
    });
    await Promise.all(writeQueue);
    common.assert.strictEqual(collected, true);
  });
  it('complete, deferred collection', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    const writeQueue = [];
    for (const block of allBlocksFlattened) {
      writeQueue.push(writer$1.put(block));
    }
    writeQueue.push(writer$1.close());
    let collected = false;
    collector(out).then(bytes => {
      collected = true;
      assertCarData(bytes);
    });
    await Promise.all(writeQueue);
    common.assert.strictEqual(collected, true);
  });
  it('complete, close after write', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    common.assert.strictEqual(typeof out[Symbol.asyncIterator], 'function');
    const collection = collector(out);
    const writeQueue = [];
    for (const block of allBlocksFlattened) {
      writeQueue.push(writer$1.put(block));
    }
    writeQueue.push(writer$1.close());
    let written = false;
    Promise.all(writeQueue).then(() => {
      written = true;
    });
    const bytes = await collection;
    common.assert.strictEqual(written, false);
    await new Promise(resolve => resolve());
    assertCarData(bytes);
  });
  it('complete, no queue', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    const collection = collector(out);
    for (const block of allBlocksFlattened) {
      await writer$1.put(block);
    }
    await writer$1.close();
    const bytes = await collection;
    assertCarData(bytes);
  });
  it('complete, slow drip', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    common.assert.strictEqual(typeof out[Symbol.asyncIterator], 'function');
    const collection = collector(out);
    for (const block of allBlocksFlattened) {
      writer$1.put(block);
      await new Promise(resolve => setTimeout(resolve, 100));
    }
    await writer$1.close();
    await new Promise(resolve => setTimeout(resolve, 100));
    const bytes = await collection;
    assertCarData(bytes);
  });
  it('complete, no queue, deferred collection', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots);
    for (const block of allBlocksFlattened) {
      writer$1.put(block);
    }
    writer$1.close();
    const collection = collector(out);
    const bytes = await collection;
    assertCarData(bytes);
  });
  it('single root', async () => {
    const {writer: writer$1, out} = writer['default'].create(roots[0]);
    const collection = collector(out);
    for (const block of allBlocksFlattened) {
      await writer$1.put(block);
    }
    await writer$1.close();
    const bytes = await collection;
    const expectedRootDef = 'a265726f6f747381d82a58250001711220f88bc853804cf294fe417e4fa83028689fcdb1b1592c5102e1474dbc200fab8b6776657273696f6e01';
    const expectedStart = (expectedRootDef.length / 2).toString(16) + expectedRootDef + '28';
    common.assert.strictEqual(toHex(bytes).substring(0, expectedStart.length), expectedStart);
  });
  it('no roots', async () => {
    const {writer: writer$1, out} = writer['default'].create();
    const collection = collector(out);
    for (const block of allBlocksFlattened) {
      await writer$1.put(block);
    }
    await writer$1.close();
    const bytes = await collection;
    const expectedRootDef = 'a265726f6f7473806776657273696f6e01';
    const expectedStart = (expectedRootDef.length / 2).toString(16) + expectedRootDef + '28';
    common.assert.strictEqual(toHex(bytes).substring(0, expectedStart.length), expectedStart);
  });
  it('bad argument for create()', () => {
    for (const arg of [
        new Uint8Array(0),
        true,
        false,
        null,
        'string',
        100,
        { obj: 'nope' },
        [false]
      ]) {
      common.assert.throws(() => writer['default'].create(arg));
    }
  });
  it('bad argument for put()', async () => {
    const {writer: writer$1} = writer['default'].create();
    for (const arg of [
        new Uint8Array(0),
        true,
        false,
        null,
        'string',
        100,
        { obj: 'nope' },
        [false]
      ]) {
      await common.assert.isRejected(writer$1.put(arg));
    }
    for (const arg of [
        true,
        false,
        null,
        'string',
        100,
        { obj: 'nope' },
        [false]
      ]) {
      await common.assert.isRejected(writer$1.put({
        bytes: new Uint8Array(0),
        cid: arg
      }));
    }
    for (const arg of [
        true,
        false,
        null,
        'string',
        100,
        { obj: 'nope' },
        [false]
      ]) {
      await common.assert.isRejected(writer$1.put({
        cid: common.rndCid,
        bytes: arg
      }));
    }
  });
  it('bad write after closed', async () => {
    const {writer: writer$1, out} = writer['default'].create();
    const collection = collector(out);
    await writer$1.put(allBlocksFlattened[0]);
    await writer$1.close();
    await common.assert.isRejected(writer$1.put(allBlocksFlattened[1]), /closed/);
    await collection;
  });
  it('bad attempt to multiple iterate', async () => {
    const {out} = writer['default'].create();
    collector(out);
    await common.assert.isRejected(collector(out), /multiple iterator/i);
  });
  it('bad attempt to multiple close', async () => {
    const {writer: writer$1, out} = writer['default'].create();
    collector(out);
    await writer$1.close();
    await common.assert.isRejected(writer$1.close(), /closed/i);
  });
});
